from Crypto.Hash import SHA3_256
from binascii import hexlify

from .utils import countries
from .merkletree import MerkleTree

def sha3(data):
    return SHA3_256.new(data).digest()

mt = MerkleTree(countries, sha3)

print(f'The root for the Merkle Tree of all countries is {hexlify(mt.root())}')

country = 'Iceland'
index, ph = mt.proof(country.encode())

print(f"The proof that {country} is in the list consists of the following {len(ph)} hashes:")
for x in ph:
    print(hexlify(x))

print(f"We only need to give these {len(ph)} hashes, along with the root, to prove that {country} is \
in the list, instead of giving {len(countries)} hashes, one for each country.")

if MerkleTree.verify(country.encode(),index,ph,mt.root(),sha3):
    print(f"{country} is in the list of countries.")
else:
    print(f"{country} is not in the list of countries.")


import unittest
from Crypto.Hash import SHA256, SHA3_256
from Crypto.Random import get_random_bytes

from .merkletree import MerkleTree

def sha256(data):
    return SHA256.new(data).digest()

def sha3(data):
    return SHA3_256.new(data).digest()

def create_tree(n, hashalg):
    data_list = [get_random_bytes(16) for _ in range(n)]
    mt = MerkleTree(data_list, hashalg)
    mt.create_tree()
    return mt

class BasicTests(unittest.TestCase):
    
    def test_different_data_different_root(self):
        mt1 = create_tree(1000, sha256)
        mt2 = create_tree(1000, sha256)
        self.assertNotEqual(mt1.root(), mt2.root())
    
    def test_different_hash_different_root(self):
        mt1 = create_tree(1000, sha256)
        mt2 = create_tree(1000, sha3)
        self.assertNotEqual(mt1.root(), mt2.root())
    
    def test_proofs_work_all_leaves(self):
        mt = create_tree(1000, sha256)
        result = []
        for x in mt.data_list:
            index, ph = mt.proof(x)
            result.append(MerkleTree.verify(x,index,ph,mt.root(),sha256))
        self.assertTrue(all(result))

class InsertUpdateTests(unittest.TestCase):

    def test_insertion_works(self):
        data_list = [b'lol',b'bam',b'boom',b'bim']
        mt1 = MerkleTree(data_list, sha256)
        mt2 = MerkleTree(data_list[:-1], sha256)
        mt2.insert(b'bim')
        self.assertEqual(mt1.tree, mt2.tree)
    
    def test_create_tree_by_insertions(self):
        data_list = [get_random_bytes(16) for _ in range(100)]
        mt1 = MerkleTree(data_list,sha256,random_padding=False,padding_byte=b'0')
        mt2 = MerkleTree([],sha256,random_padding=False,padding_byte=b'0')
        for x in data_list:
            mt2.insert(x)
        self.assertEqual(mt1.tree, mt2.tree)

    def test_update_works(self):
        data_list1 = [get_random_bytes(16) for _ in range(100)]
        for i in range(len(data_list1)):
            mt1 = MerkleTree(data_list1,sha256,random_padding=False,padding_byte=b'0')
            data_list2 = data_list1[:i] + [b'foo'] + data_list1[i+1:]
            mt2 = MerkleTree(data_list2,sha256,random_padding=False,padding_byte=b'0')
            mt1.update(i, b'foo')
            with self.subTest(i=i):
                self.assertEqual(mt1.tree, mt2.tree)
    
    def test_create_tree_by_updates(self):
        data_list = [get_random_bytes(16) for _ in range(100)]
        mt1 = MerkleTree(data_list,sha256,random_padding=False,padding_byte=b'0')
        mt2 = MerkleTree([get_random_bytes(16) for _ in range(100)],sha256,random_padding=False,padding_byte=b'0')
        for i, x in enumerate(data_list):
            mt2.update(i,x)
        self.assertEqual(mt1.tree, mt2.tree)

if __name__ == '__main__':
    unittest.main()


from Crypto.Random import get_random_bytes
from Crypto.Hash import SHA256
from math import log2, ceil

from .utils import is_power2

class MerkleTree:
    def __init__(self, data_list, hashalg, random_padding=True, padding_byte=None):
        self.data_list = data_list
        self.hashdata = [hashalg(x) for x in data_list]
        self.hashalg = hashalg
        self.depth = ceil(log2(len(data_list))) if len(data_list) else 0
        self.random_padding = random_padding
        self.padding_byte = padding_byte
        if random_padding:
            self.padding = (2**self.depth - len(self.data_list))*[hashalg(get_random_bytes(16))]
        else:
            self.padding = (2**self.depth - len(self.data_list))*[hashalg(padding_byte)]
        self.tree = {}
        self.create_tree()
    
    def create_tree(self):
        self.tree[self.depth] = self.hashdata + self.padding
        for i in range(self.depth-1,-1,-1):
            level = self.tree[i+1]
            self.tree[i] = []
            for j in range(0,len(level),2):
                left = level[j]
                right = level[j+1]
                self.tree[i].append(self.hashalg(left + right))

    def root(self):
        return self.tree[0][0] 
    
    def update(self, index, data):
        self.data_list[index] = data
        self.hashdata[index] = self.hashalg(data)
        self._update_tree(index)

    
    def _update_tree(self, index):
        self.tree[self.depth][index] = self.hashdata[index]
        for i in range(self.depth-1,-1,-1):
            if index%2:
                self.tree[i][index//2] = self.hashalg(
                    self.tree[i+1][index-1] + self.tree[i+1][index]
                )
            else:
                self.tree[i][index//2] = self.hashalg(
                    self.tree[i+1][index] + self.tree[i+1][index+1]
                )
            index //= 2

    
    def insert(self, data):
        if not is_power2(len(self.data_list)):    
            index = len(self.data_list)
            self.data_list.append(data)
            self.hashdata.append(self.hashalg(data))
            self.padding = self.padding[1:]

            self._update_tree(index)
        else:
            self.depth += 1
            index = len(self.data_list)
            self.data_list.append(data)
            self.hashdata.append(self.hashalg(data))
            if self.random_padding:
                self.padding = (2**self.depth - len(self.data_list))*[self.hashalg(get_random_bytes(16))]
            else:
                self.padding = (2**self.depth - len(self.data_list))*[self.hashalg(self.padding_byte)]
            
            self.create_tree()

    def proof(self, data):
        original_index = index = self.data_list.index(data)
        proof_hashes = []
        for i in range(self.depth,0,-1):
            complementary_index = index-1 if index%2 else index+1
            proof_hashes.append(
                self.tree[i][complementary_index]
            )
            index //= 2
        
        return original_index, proof_hashes
    
    @staticmethod
    def verify(data, index, proof_hashes, root, hashalg):
        h = hashalg(data)
        for x in proof_hashes:
            if index%2:
                h = hashalg(x + h)
            else:
                h = hashalg(h + x)
            index //= 2
        
        return h == root


def is_power2(num):
    'states if a number is a power of two'
    return num != 0 and ((num & (num - 1)) == 0)

countries = [
    b'Afghanistan',
    b'Albania',
    b'Algeria',
    b'Andorra',
    b'Angola',
    b'Antigua & Deps',
    b'Argentina',
    b'Armenia',
    b'Australia',
    b'Austria',
    b'Azerbaijan',
    b'Bahamas',
    b'Bahrain',
    b'Bangladesh',
    b'Barbados',
    b'Belarus',
    b'Belgium',
    b'Belize',
    b'Benin',
    b'Bhutan',
    b'Bolivia',
    b'Bosnia Herzegovina',
    b'Botswana',
    b'Brazil',
    b'Brunei',
    b'Bulgaria',
    b'Burkina',
    b'Burundi',
    b'Cambodia',
    b'Cameroon',
    b'Canada',
    b'Cape Verde',
    b'Central African Rep',
    b'Chad',
    b'Chile',
    b'China',
    b'Colombia',
    b'Comoros',
    b'Congo',
    b'Congo',
    b'Costa Rica',
    b'Croatia',
    b'Cuba',
    b'Cyprus',
    b'Czech Republic',
    b'Denmark',
    b'Djibouti',
    b'Dominica',
    b'Dominican Republic',
    b'East Timor',
    b'Ecuador',
    b'Egypt',
    b'El Salvador',
    b'Equatorial Guinea',
    b'Eritrea',
    b'Estonia',
    b'Ethiopia',
    b'Fiji',
    b'Finland',
    b'France',
    b'Gabon',
    b'Gambia',
    b'Georgia',
    b'Germany',
    b'Ghana',
    b'Greece',
    b'Grenada',
    b'Guatemala',
    b'Guinea',
    b'Guinea-Bissau',
    b'Guyana',
    b'Haiti',
    b'Honduras',
    b'Hungary',
    b'Iceland',
    b'India',
    b'Indonesia',
    b'Iran',
    b'Iraq',
    b'Ireland',
    b'Israel',
    b'Italy',
    b'Ivory Coast',
    b'Jamaica',
    b'Japan',
    b'Jordan',
    b'Kazakhstan',
    b'Kenya',
    b'Kiribati',
    b'Korea North',
    b'Korea South',
    b'Kosovo',
    b'Kuwait',
    b'Kyrgyzstan',
    b'Laos',
    b'Latvia',
    b'Lebanon',
    b'Lesotho',
    b'Liberia',
    b'Libya',
    b'Liechtenstein',
    b'Lithuania',
    b'Luxembourg',
    b'Macedonia',
    b'Madagascar',
    b'Malawi',
    b'Malaysia',
    b'Maldives',
    b'Mali',
    b'Malta',
    b'Marshall Islands',
    b'Mauritania',
    b'Mauritius',
    b'Mexico',
    b'Micronesia',
    b'Moldova',
    b'Monaco',
    b'Mongolia',
    b'Montenegro',
    b'Morocco',
    b'Mozambique',
    b'Myanmar',
    b'Namibia',
    b'Nauru',
    b'Nepal',
    b'Netherlands',
    b'New Zealand',
    b'Nicaragua',
    b'Niger',
    b'Nigeria',
    b'Norway',
    b'Oman',
    b'Pakistan',
    b'Palau',
    b'Panama',
    b'Papua New Guinea',
    b'Paraguay',
    b'Peru',
    b'Philippines',
    b'Poland',
    b'Portugal',
    b'Qatar',
    b'Romania',
    b'Russian Federation',
    b'Rwanda',
    b'St Kitts & Nevis',
    b'St Lucia',
    b'Saint Vincent & the Grenadines',
    b'Samoa',
    b'San Marino',
    b'Sao Tome & Principe',
    b'Saudi Arabia',
    b'Senegal',
    b'Serbia',
    b'Seychelles',
    b'Sierra Leone',
    b'Singapore',
    b'Slovakia',
    b'Slovenia',
    b'Solomon Islands',
    b'Somalia',
    b'South Africa',
    b'South Sudan',
    b'Spain',
    b'Sri Lanka',
    b'Sudan',
    b'Suriname',
    b'Swaziland',
    b'Sweden',
    b'Switzerland',
    b'Syria',
    b'Taiwan',
    b'Tajikistan',
    b'Tanzania',
    b'Thailand',
    b'Togo',
    b'Tonga',
    b'Trinidad & Tobago',
    b'Tunisia',
    b'Turkey',
    b'Turkmenistan',
    b'Tuvalu',
    b'Uganda',
    b'Ukraine',
    b'United Arab Emirates',
    b'United Kingdom',
    b'United States',
    b'Uruguay',
    b'Uzbekistan',
    b'Vanuatu',
    b'Vatican City',
    b'Venezuela',
    b'Vietnam',
    b'Yemen',
    b'Zambia',
    b'Zimbabwe'
]
